/*

	EFIPW (C) 2009 by Paul Makowski (my.hndl@gmail.com)
	http://code.google.com/p/efipw/
	http://paulmakowski.blogspot.com/
	
	License:	
	GPLv3
	
	Warranty:	
	Unless otherwise stated, this software is distributed 
	in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the 
	implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. THERE IS 
	NO WARRANTY FOR THE SOFTWARE, TO THE EXTENT PERMITTED BY APPLICABLE LAW. EXCEPT 
	WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS PROVIDE THE SOFTWARE "AS IS" 
	WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED
	TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. 
	THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE SOFTWARE IS WITH YOU. SHOULD 
	THE SOFTWARE PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR 
	OR CORRECTION. IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING 
	WILL ANY COPYRIGHT HOLDER, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY GENERAL, 
	SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OR INABILITY TO 
	USE THE SOFTWARE (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED 
	INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE OF THE SOFTWARE 
	TO OPERATE WITH ANY OTHER PROGRAMS), EVEN IF SUCH HOLDER HAS BEEN ADVISED OF THE 
	POSSIBILITY OF SUCH DAMAGES. 

	About:
	EFIPW is a tool to modify and/or decode firmware passwords stored on EFI-based Apple 
	computers in nvram.  It may be useful for both lab administrators and pen testers.
	It was inspired by the OPFW tool.  Requires root permissions.
	
	If you find this software useful, I would appreciate it if you let me know, but 
	you're certainly not required to.
	
	Usage (as unprivileged user):
	To reveal current firmware password: sudo ./EFIPW -d
	To set a new password: sudo ./EFIPW -p "newpassword" (-m "newmode")
		Valid modes: none, command (default), full
		
	Todo:
	- better arg parsing
	- garbage collection
	- don't be so lazy with the encoding of new passwords
	- test maximum length controls

*/

#define VERSION	"0.1"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <math.h>

int i;

void usage (void)
{
	printf("\nUsage (as unprivileged user):\n");
	printf("To reveal current firmware password: sudo ./EFIPW -d\n");
	printf("To set a new password: sudo ./EFIPW -p \"newpassword\" (-m \"newmode\")\n");
	printf("\tValid modes: none, command (default), full\n\n");
	exit(1);
}


int revealPW (void)
{
	FILE *f;
	char temp[785]; // 3 * 255 + 19 + 1 (%ab per char plus 'security-password\t')
	char encodedBinary[2041] = ""; // 8 * 255 (8 bit ASCII encoding)
	char decodedBinary[2041] = "";

	f=popen("nvram -p | grep security-password", "r");
	fgets(temp, 784, f);
	pclose(f);

	/* pull out the part of the output we're concerned with */
	char *encodedHex = strchr(temp, '\t');
	encodedHex++; // now encodedHex points to the beginning of encoded password
	memset(encodedHex+strlen(encodedHex)-1, 0, 1); // kill the newline
	
	char *ptr = encodedHex;
	
	/* convert encodedHex to binary string - there's probably a much better way to do this */
	for(i=0; i < strlen(encodedHex)-1; i++)
	{
		ptr++;
		switch(*ptr)
		{
			case 37: break; // %
			case '0': strcat(encodedBinary, "0000"); break;
			case '1': strcat(encodedBinary, "0001"); break;
			case '2': strcat(encodedBinary, "0010"); break;
			case '3': strcat(encodedBinary, "0011"); break;
			case '4': strcat(encodedBinary, "0100"); break;
			case '5': strcat(encodedBinary, "0101"); break;
			case '6': strcat(encodedBinary, "0110"); break;
			case '7': strcat(encodedBinary, "0111"); break;
			case '8': strcat(encodedBinary, "1000"); break;
			case '9': strcat(encodedBinary, "1001"); break;
			case 'a': strcat(encodedBinary, "1010"); break;
			case 'b': strcat(encodedBinary, "1011"); break;
			case 'c': strcat(encodedBinary, "1100"); break;
			case 'd': strcat(encodedBinary, "1101"); break;
			case 'e': strcat(encodedBinary, "1110"); break;
			case 'f': strcat(encodedBinary, "1111"); break;
			
			default: printf("Invalid hex value '%c'.  Dying...\n", ptr); return 1;
		}
	}
	
	/* apply the magic formula */
	int invert=1; ptr = encodedBinary;
	for(i=0; i<strlen(encodedBinary); i++)
	{
		if(invert)
		{
			if(*ptr == '1')
				strcat(decodedBinary, "0");
			else
				strcat(decodedBinary, "1");
			invert=0;
		}
		else
		{
			strncat(decodedBinary, ptr, 1);
			invert=1;
		}
		ptr++;
	}
	
	/* interpret decoded binary as ACSII, print it */
	char bit; ptr = decodedBinary;
	for (i=0; i<strlen(decodedBinary); i+=8)
	{
		/* pull out 8 bits of decodedBinary */
		snprintf(temp, 9, ptr+i, "%s");
		
		/* convert to decimal */
		int j; int tempDec=0;
		for (j=0; j<8; j++)
		{
			bit = *(ptr+i+j);
			if (bit == '1')
				tempDec+=pow(2, 7-j);
		}
		printf("%c", tempDec);
	}
	printf("\n");
	
	/* temp, encodedBinary, decodedBinary being leaked... should probably fix at some point */
	return 0;
}

void updatePW (char *newPass, char *newMode)
{
	/* set default mode if unset -- FIX THIS */
	if (newMode == NULL)
		newMode = "command";
	if (newPass == NULL)
		newPass = "";
	
	/* make sure mode is valid */
	if(strcmp(newMode, "none") != 0 && strcmp(newMode, "command") != 0 && strcmp(newMode, "full") != 0)
		usage();
		
	/* make sure new password is valid */
	if (strcmp(newPass, "none") == 0)
		printf("warning: setting password to \"none\"\n");
		
	/* make sure new password is of appropriate length */
	if (strlen(newPass) > 255)
	{
		printf("New password is too long.  Maximum length: 255 characters.  Dying...\n");
		exit(1);
	}
	
	/* I GOT REALLY LAZY HERE  */
	char temp[766] = ""; char currChar;
	for (i=0; i<strlen(newPass); i++)
	{
		currChar = *(newPass+i);
		switch(currChar)
		{
			case 32: strcat(temp, "%8a"); break; case 33: strcat(temp, "%8b"); break; case 34: strcat(temp, "%88"); break; case 35: strcat(temp, "%89"); break;
			case 36: strcat(temp, "%8e"); break; case 37: strcat(temp, "%8f"); break; case 38: strcat(temp, "%8c"); break; case 39: strcat(temp, "%8d"); break;
			case 40: strcat(temp, "%82"); break; case 41: strcat(temp, "%83"); break; case 42: strcat(temp, "%80"); break; case 43: strcat(temp, "%81"); break; 
			case 44: strcat(temp, "%86"); break; case 45: strcat(temp, "%87"); break; case 46: strcat(temp, "%84"); break; case 47: strcat(temp, "%85"); break;
			
			case 48: strcat(temp, "%9a"); break; case 49: strcat(temp, "%9b"); break; case 50: strcat(temp, "%98"); break; case 51: strcat(temp, "%99"); break;
			case 52: strcat(temp, "%9e"); break; case 53: strcat(temp, "%9f"); break; case 54: strcat(temp, "%9c"); break; case 55: strcat(temp, "%9d"); break;
			case 56: strcat(temp, "%92"); break; case 57: strcat(temp, "%93"); break; case 58: strcat(temp, "%90"); break; case 59: strcat(temp, "%91"); break;
			case 60: strcat(temp, "%96"); break; case 61: strcat(temp, "%97"); break; case 62: strcat(temp, "%94"); break; case 63: strcat(temp, "%95"); break;
			
			case 64: strcat(temp, "%ea"); break; case 65: strcat(temp, "%eb"); break; case 66: strcat(temp, "%e8"); break; case 67: strcat(temp, "%e9"); break;
			case 68: strcat(temp, "%ee"); break; case 69: strcat(temp, "%ef"); break; case 70: strcat(temp, "%ec"); break; case 71: strcat(temp, "%ed"); break;
			case 72: strcat(temp, "%e2"); break; case 73: strcat(temp, "%e3"); break; case 74: strcat(temp, "%e0"); break; case 75: strcat(temp, "%e1"); break;
			case 76: strcat(temp, "%e6"); break; case 77: strcat(temp, "%e7"); break; case 78: strcat(temp, "%e4"); break; case 79: strcat(temp, "%e5"); break;
			
			case 80: strcat(temp, "%fa"); break; case 81: strcat(temp, "%fb"); break; case 82: strcat(temp, "%f8"); break; case 83: strcat(temp, "%f9"); break;
			case 84: strcat(temp, "%fe"); break; case 85: strcat(temp, "%ff"); break; case 86: strcat(temp, "%fc"); break; case 87: strcat(temp, "%fd"); break; 
			case 88: strcat(temp, "%f2"); break; case 89: strcat(temp, "%f3"); break; case 90: strcat(temp, "%f0"); break; case 91: strcat(temp, "%f1"); break;
			case 92: strcat(temp, "%f6"); break; case 93: strcat(temp, "%f7"); break; case 94: strcat(temp, "%f4"); break; case 95: strcat(temp, "%f5"); break;
			
			case 96: strcat(temp, "%ca"); break; case 97: strcat(temp, "%cb"); break; case 98: strcat(temp, "%c8"); break; case 99: strcat(temp, "%c9"); break;
			case 100: strcat(temp, "%ce"); break; case 101: strcat(temp, "%cf"); break; case 102: strcat(temp, "%cc"); break; case 103: strcat(temp, "%cd"); break;
			case 104: strcat(temp, "%c2"); break; case 105: strcat(temp, "%c3"); break; case 106: strcat(temp, "%c0"); break; case 107: strcat(temp, "%c1"); break;
			case 108: strcat(temp, "%c6"); break; case 109: strcat(temp, "%c7"); break;  case 110: strcat(temp, "%c4"); break; case 111: strcat(temp, "%c5"); break;
			
			case 112: strcat(temp, "%da"); break; case 113: strcat(temp, "%db"); break; case 114: strcat(temp, "%d8"); break; case 115: strcat(temp, "%d9"); break;
			case 116: strcat(temp, "%de"); break; case 117: strcat(temp, "%df"); break; case 118: strcat(temp, "%dc"); break; case 119: strcat(temp, "%dd"); break;
			case 120: strcat(temp, "%d2"); break; case 121: strcat(temp, "%d3"); break; case 122: strcat(temp, "%d0"); break; case 123: strcat(temp, "%d1"); break;
			case 124: strcat(temp, "%d6"); break; case 125: strcat(temp, "%d7"); break; case 126: strcat(temp, "%d4"); break; case 127: strcat(temp, "%d5"); break;
			
			default: printf("Invalid character w/ ASCII decimal value: '%d' in new password.  Passwords must be all ASCII between 32 and 127 (decimal).  Dying...\n", currChar); exit(1);
		}
	}
		
	char command[813];
	snprintf(command, 812, "nvram security-password=%s security-mode=%s", &temp, newMode);
	system(command);
	
	return;
}

int main (int argc, char **argv) 
{
	/* check if its run as root */
	if (geteuid() != 0) { printf("Must be run as root.\n"); exit(1); }
	
	/* defaults... overwritten with args */
	int decryptMode = 0; char *newPass = NULL; char *newMode = NULL;
	
	/* parse the args and set flags */
	while ((i = getopt (argc, argv, "dp:m:")) != -1)
	{
		switch (i)
		{
			case 'd':
				decryptMode = 1;
				break;
			case 'p':
				newPass = optarg;
				break;
			case 'm':
				newMode = optarg;
				break;
			case '?':
				usage();
				return 1;
			default:
				abort ();
		}
	}
	
	/* call appropriate functions */
	if (decryptMode) { revealPW(); }
	else { updatePW(newPass, newMode); }
	return 0;
}
